{*********************************************************}
{*          Web server application components            *}
{*                     version 2.0                       *}
{*         Copyright (c) 2000-2002  S.Perevoznyk         *}
{*                 All rights reserved.                  *}
{*********************************************************}
unit psvWeb;

interface

uses
  Windows, Messages, SysUtils, Classes,
  HTTPApp,
  {$IFDEF VER140} HTTPProd, {$ENDIF} CopyPrsr;

type

  TpsvParseScriptEvent = procedure(Sender : TObject; var ScriptText : string; var Handled : boolean) of object;

  TpsvScriptItem = class(TCollectionItem)
  private
    FScriptName : string;
    FLines      : TStringList;
    procedure SetLines(Value : TStringList);
  public
    constructor Create(Collection : TCollection); override;
    destructor  Destroy; override;
    procedure   Assign(Source: TPersistent); override;
  published
    property ScriptName : string read FScriptName write FScriptName;
    property Lines      : TStringList read FLines write SetLines;
  end;

  TpsvScriptCollection = class(TCollection)
  private
     FOwner : TComponent;
     function    GetItem(Index: Integer): TpsvScriptItem;
     procedure   SetItem(Index: Integer; Value: TpsvScriptItem);
   protected
     function    GetOwner: TPersistent; override;
   public
     constructor Create(AOwner : TComponent);
     function    Add: TpsvScriptItem;
     property    Items[Index: Integer]: TpsvScriptItem read GetItem write SetItem; default;
   end;


  TpsvTagItem = class(TCollectionItem)
  private
    FHTMLFile : TFileName;
    FTagName  : string;
    FTagValue : TStringList;
    FOnTag    : THTMLTagEvent;
    procedure SetHTMLFile(Value : TFileName);
    procedure SetTagValue(Value : TStringList);
  protected
    procedure DoTagEvent(Tag: TTag; const TagString: string; TagParams: TStrings;
                          var ReplaceText: string); dynamic;
  public
    procedure Assign(Source : TPersistent); override;
    constructor Create(Collection : TCollection); override;
    destructor  Destroy; override;
  published
    property TagName : string read FTagName  write FTagName;
    property Value   : TStringList read FTagValue write SetTagValue;
    property HTMLFile: TFileName read FHTMLFile write SetHTMLFile;
    property OnHTMLTag :THTMLTagEvent read FOnTag write FOnTag;
  end;

  TpsvTagCollection = class(TCollection)
  private
     FOwner : TComponent;
     function    GetItem(Index: Integer): TpsvTagItem;
     procedure   SetItem(Index: Integer; Value: TpsvTagItem);
   protected
     function    GetOwner: TPersistent; override;
   public
     constructor Create(AOwner : TComponent);
     function    Add: TpsvTagItem;
     property    Items[Index: Integer]: TpsvTagItem read GetItem write SetItem; default;
   end;


  TpsvCustomTagLib = class(TComponent)
  private
    FTags : TpsvTagCollection;
    procedure SetTags(Value : TpsvTagCollection);
  public
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy; override;
    property    Tags : TpsvTagCollection read FTags write SetTags;
  end;

  TpsvTagLib = class(TpsvCustomTagLib)
  published
    property Tags;
  end;
    
  TpsvScriptLib = class(TComponent)
  private
    FScripts : TpsvScriptCollection;
    procedure SetScripts(Value : TpsvScriptCollection);
  public
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy; override;
  published
    property    Scripts : TpsvScriptCollection read FScripts write SetScripts;
  end;

  TpsvPageProducer = class(TPageProducer)
  private
     FOnParseScript : TpsvParseScriptEvent;
     FScriptLib : TpsvScriptLib;
     FTagLibs   : TList;
     FEnabled   : boolean;
     FHandleStandardTags : boolean;
     FBeforeTagEvent : THTMLTagEvent;
     FBeforeProduceContent : TNotifyEvent;
     FScriptBlocks : TStringList;
     procedure SetScriptLib( Value : TpsvScriptLib);
  protected
     procedure Notification(AComponent: TComponent; Operation: TOperation); override;
     procedure DoTagEvent(Tag: TTag; const TagString: string; TagParams: TStrings;
       var ReplaceText: string); override;
     procedure DoParseScript(AScript : string; var AHandled : boolean); virtual;
  public
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy; override;
    property    Enabled : boolean read FEnabled write FEnabled;
    function    ContentFromStream(Stream: TStream): string; override;
    procedure   RegisterLibrary(TagLib : TpsvCustomTagLib);
  published
    property ScriptLib : TpsvScriptLib read FScriptLib write SetScriptLib;
    property HandleStandardTags : boolean read FHandleStandardTags write FHandleStandardTags default true;
    property BeforeHTMLTag : THTMLTagEvent read FBeforeTagEvent write FBeforeTagEvent;
    property BeforeProduceContent : TNotifyEvent read FBeforeProduceContent write FBeforeProduceContent;
    property OnParseScript : TpsvParseScriptEvent read FOnParseScript write FOnParseScript;
  end;


procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('FastWeb', [TpsvPageProducer, TpsvScriptLib, TpsvTagLib]);
end;


procedure  TpsvTagItem.Assign(Source : TPersistent);
begin
  if Source is TpsvTagItem then
  begin
    FTagName    := TpsvTagItem(Source).TagName;
    FTagValue   := TpsvTagItem(Source).Value;
    Exit;
  end;
  inherited Assign(Source);
end;

constructor TpsvTagItem.Create(Collection: TCollection);
begin
  inherited;
  FTagValue := TStringList.Create;
end;

destructor TpsvTagItem.Destroy;
begin
  inherited;
  FTagValue.Free;
end;

procedure TpsvTagItem.DoTagEvent(Tag: TTag; const TagString: string; TagParams: TStrings;
      var ReplaceText: string);
var
  InStream: TStream;

begin
  InStream := Nil;
  
  if FHTMLFile <> '' then
    InStream := TFileStream.Create(FHTMLFile, fmOpenRead + fmShareDenyWrite)
      else
        if FTagValue.Text <> '' then
          InStream := TStringStream.Create(FTagValue.Text);

  if InStream <> nil then
  try
    SetLength(ReplaceText,InStream.Size);
    InStream.Read(ReplaceText[1],InStream.Size);
  finally
    InStream.Free;
  end
    else
       if Assigned(FOnTag) then
          FOnTag(Self, Tag, TagString, TagParams, ReplaceText);
end;



procedure TpsvTagItem.SetHTMLFile(Value: TFileName);
begin
  if CompareText(FHTMLFile, Value) <> 0 then
  begin
    FTagValue.Clear;
    FHTMLFile := Value;
  end;
end;

procedure TpsvTagItem.SetTagValue(Value: TStringList);
begin
  FTagValue.Assign(Value);
end;


constructor TpsvScriptItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FLines := TstringList.Create;
end;


destructor TpsvScriptItem.Destroy;
begin
 FLines.Free;
 inherited Destroy;
end;

procedure TpsvScriptItem.SetLines(Value : TStringList);
begin
  FLines.Assign(Value);
end;

procedure TpsvScriptItem.Assign(Source: TPersistent);
begin
  if Source is TpsvScriptItem then
  begin
    FScriptName    := TpsvScriptItem(Source).ScriptName;
    FLines         := TpsvScriptItem(Source).Lines;
    Exit;
  end;
  inherited Assign(Source);
end;

constructor TpsvScriptCollection.Create(AOwner : TComponent);
begin
  inherited Create(TpsvScriptItem);
  FOwner := AOwner;
end;

function   TpsvScriptCollection.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

function TpsvScriptCollection.GetItem(Index: Integer): TpsvScriptItem;
begin
  Result := TpsvScriptItem(inherited GetItem(Index));
end;

procedure TpsvScriptCollection.SetItem(Index: Integer; Value: TpsvScriptItem);
begin
  inherited SetItem(Index, Value)
end;

function TpsvScriptCollection.Add: TpsvScriptItem;
begin
  Result := TpsvScriptItem(inherited Add);
end;


{TpsvTagCollection class}
constructor TpsvTagCollection.Create(AOwner : TComponent);
begin
  inherited Create(TpsvTagItem);
  FOwner := AOwner;
end;

function   TpsvTagCollection.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

function TpsvTagCollection.GetItem(Index: Integer): TpsvTagItem;
begin
  Result := TpsvTagItem(inherited GetItem(Index));
end;

procedure TpsvTagCollection.SetItem(Index: Integer; Value: TpsvTagItem);
begin
  inherited SetItem(Index, Value)
end;

function TpsvTagCollection.Add: TpsvTagItem;
begin
  Result := TpsvTagItem(inherited Add);
end;


{ TpsvScriptLib }

constructor TpsvScriptLib.Create(AOwner: TComponent);
begin
  inherited;
  FScripts := TpsvScriptCollection.Create(Self);
end;

destructor TpsvScriptLib.Destroy;
begin
  FScripts.Free;
  inherited;
end;


procedure TpsvScriptLib.SetScripts(Value: TpsvScriptCollection);
begin
  FScripts.Assign(Value);
end;

{ TpsvPageProducer }

constructor TpsvPageProducer.Create(AOwner : TComponent);
begin
  inherited;
  FEnabled := true;
  FHandleStandardTags := true;
  FTagLibs := TList.Create;
  FScriptBlocks := TStringList.Create;
end;


destructor TpsvPageProducer.Destroy;
begin
  FTagLibs.Free;
  FScriptBlocks.Free;
  inherited;
end;

procedure TpsvPageProducer.RegisterLibrary(TagLib : TpsvCustomTagLib);
begin
  if Assigned(TagLib) then
   begin
     FTagLibs.Add(TagLib);
   end;
end;

procedure TpsvPageProducer.DoParseScript(AScript : string; var AHandled : boolean);
begin
  AHandled := false;
  if Assigned(FOnParseScript) then
    FOnParseScript(Self, AScript, AHandled);
end;

function TpsvPageProducer.ContentFromStream(Stream: TStream): string;
var
  Parser: TCopyParser;
  OutStream: TStringStream;
  ResultStream : TStringStream;
  ParamStr, ReplaceStr, TokenStr: string;
  ParamList: TStringList;
  RealParamStr : string;
  SavePos, SaveSize : integer;
  ScriptNum : integer;
  Handled : boolean;

procedure DoTokenWork;
var AsStandard : boolean;
    PrevToken  : char;
begin
  with Parser do
   begin
     PrevToken := SkipToken(False);
     if PrevToken = '#' then
      begin
       SkipToken(False);
       AsStandard := false;
      end
       else
        AsStandard := true;
     TokenStr := TokenString;
     RealParamStr := SkipToToken('>');
     ParamStr := TrimLeft(TrimRight(RealParamStr));
     ParamList := TStringList.Create;
     try
      ExtractHeaderFields([' '], [' '], PChar(ParamStr), ParamList, false, StripParamQuotes);

      ReplaceStr := HandleTag(TokenStr, ParamList);
      if AsStandard then
       begin
         if ReplaceStr = EmptyStr then
          ReplaceStr := '<'+TokenStr+RealParamStr + '>';
       end;
      if FEnabled then
       ResultStream.WriteString(ReplaceStr);
      finally
       ParamList.Free;
      end;
      SkipToken(True);
   end;
end;

begin
  if Assigned(FBeforeProduceContent) then
   FBeforeProduceContent(Self);
  OutStream := TStringStream.Create('');
  ResultStream := TStringStream.Create('');
  FScriptBlocks.Clear;

  //first cycle
 {$IFDEF VER140}
 if ScriptEngine = '' then
  begin
 {$ENDIF}
  Parser := TCopyParser.Create(Stream,OutStream);
  try
   with Parser do
    begin
      CopyTokenToOutput;
      While True do
        begin
          while not (Token in [toEof, '<']) do
              NextToken;
          if Token = toEOF then Break;
          if Token = '<' then
           begin
             SavePos := OutStream.Position;
             SaveSize := OutStream.Size;
             while true do
             begin
               NextToken;
               if Token = '%' then
                begin
                  while true do
                   begin
                      NextToken;
                      if Token = '%' then
                        begin
                          if NextToken = '>' then
                          break;
                        end
                          else
                            if Token = toEOF then
                               break;
                    end;
                end
                  else
                    if Token = '>' then
                      break
                        else
                          if Token = toEOF then
                             break;
             end;
             TokenStr := Copy(OutStream.DataString,SavePos,OutStream.Position - SavePos + 1);
             OutStream.Position := SavePos-1;
             OutStream.Size := SaveSize-1;
             if Pos('%',TokenStr) <> 2 then
              OutStream.WriteString(TokenStr)
               else
                begin
                  Delete(TokenStr,1,2);
                  SetLength(TokenStr,Length(TokenStr)-2);
                  TokenStr := Trim(TokenStr);
                  DoParseScript(TokenStr, Handled);
                  if not Handled then
                    begin
                      ScriptNum := FScriptBlocks.Add(TokenStr);
                      OutStream.WriteString(Format('<#PSCRIPT NAME="%d"',[ScriptNum]));
                    end;
                end;
           end;
        end;
    end;
  finally
   Parser.Free;
  end;
  {$IFDEF VER140}
   end
    else
      OutStream.CopyFrom(Stream,Stream.Size);
  {$ENDIF}

  OutStream.Position := 0;
  //second cycle
  try
    Parser := TCopyParser.Create(OutStream, ResultStream);
    with Parser do
    try
      while True do
      begin
        while not (Token in [toEof, '<']) do
        begin
          if FEnabled then
            CopyTokenToOutput;
            SkipToken(True);
        end;
        if Token = toEOF then Break;
        if Token = '<' then
           DoTokenWork
             else
                begin
                  if FEnabled then
                  begin
                    ResultStream.WriteString('<');
                    CopyTokenToOutput;
                  end;
                   SkipToken(True);
               end;
        end;
    finally
      Parser.Free;
    end;
    Result := ResultStream.DataString;
  finally
    ResultStream.Free;
  end;
end;


procedure TpsvPageProducer.SetScriptLib(Value: TpsvScriptLib);
begin
  if FScriptLib <> Value then
  begin
    if Value <> nil then Value.FreeNotification(Self);
    FScriptLib := Value;
  end;
end;

procedure TpsvPageProducer.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = ScriptLib) then
    FScriptLib := nil;

  if (Operation = opInsert) and (AComponent is TpsvScriptLib) then
    FScriptLib := TpsvScriptLib(AComponent);

end;

procedure TpsvPageProducer.DoTagEvent(Tag: TTag; const TagString: string;
  TagParams: TStrings; var ReplaceText: string);
var
  i  : integer;
  St : string;
  LibCount : integer;

begin
  if Assigned(FBeforeTagEvent) then
   FBeforeTagEvent(Self, Tag, TagString, TagParams,ReplaceText);
  for LibCount := 0 to FTagLibs.Count - 1 do
  begin
    if Assigned(FTagLibs[LibCount]) then
    for i := 0 to TpsvCustomTagLib(FTagLibs[LibCount]).Tags.Count -1 do
     begin
       if ( AnsiCompareText(TagString, TpsvCustomTagLib(FTagLibs[LibCount]).Tags[I].TagName) = 0) then
        begin
          TpsvCustomTagLib(FTagLibs[LibCount]).Tags[I].DoTagEvent(Tag,TagString,TagParams,ReplaceText);
          Exit;
        end;
     end;
  end;

 if  Assigned(FScriptLib) then
  begin
    if ( AnsiCompareText(TagString,'JAVALIB') = 0)  then
     begin
       st := TagParams.Values['SCRIPTNAME'];
       for i := 0 to FScriptLib.Scripts.Count -1 do
        begin
          if ( AnsiCompareText(St,FScriptLib.Scripts[I].ScriptName) = 0 ) then
           begin
             ReplaceText := FScriptLib.Scripts[I].Lines.Text;
             Exit;
           end;
        end;
     end;
  end;

  //Additional Tags


  if AnsiCompareText(TagString,'DATE') = 0 then
   begin
     ReplaceText :=  DateToStr(Date);
     Exit;
   end;

  if AnsiCompareText(TagString, 'TIME') = 0 then
   begin
     ReplaceText := TimeToStr(Time);
     Exit;
   end;

  if AnsiCompareText(TagString,'NOW') = 0 then
   begin
     ReplaceText := DateTimeToStr(Now);
     Exit;
   end;

  if AnsiCompareText(TagString,'DAY') = 0 then
   begin
     if TagParams.Values['DATE'] = EmptyStr then
       ReplaceText := LongDayNames[DayOfWeek(Now)]
         else
           ReplaceText := LongDayNames[DayOfWeek(StrToDate(TagParams.Values['DATE']))];
      Exit;
    end;


  inherited DoTagEvent(Tag, TagString, TagParams, ReplaceText);
end;


{ TpsvCustomTagLib }

constructor TpsvCustomTagLib.Create(AOwner: TComponent);
begin
  inherited;
  FTags := TpsvTagCollection.Create(Self);
end;

destructor TpsvCustomTagLib.Destroy;
begin
  FTags.Free;
  inherited;
end;

procedure TpsvCustomTagLib.SetTags(Value: TpsvTagCollection);
begin
  FTags.Assign(Value);
end;



end.
